/*******************************************************************************
 * Copyright (c) 2007-2008 SAS Institute Inc., ILOG S.A.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     SAS Institute Inc. - initial API and implementation
 *     ILOG S.A. - initial API and implementation
 *******************************************************************************/
package org.eclipse.albireo.core;

import java.awt.EventQueue;
import java.io.PrintStream;

import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import org.eclipse.albireo.internal.Platform;

/**
 * This class deals with the customization of the look&amp;feel
 * of Swing components embedded inside SWT.
 * <p>
 * It is customizable through the "replaceable singleton" design pattern.
 */
public class LookAndFeelHandler {

    // ========================================================================
    // Accessors

    /**
     * This look&amp;feel choice denotes the default Swing look&amp;feel.
     * It may be platform dependent.
     */
    public static final String LAFChoiceSwingDefault = "swing.defaultlaf";

    /**
     * This look&amp;feel choice denotes the cross-platform Swing
     * look&amp;feel.
     * It is not platform dependent.
     * @see javax.swing.UIManager#getCrossPlatformLookAndFeelClassName()
     */
    public static final String LAFChoiceCrossPlatform = "swing.crossplatformlaf";

    /**
     * This look&amp;feel choice denotes the Swing look&amp;feel
     * that best approximates the system look&amp;feel.
     * It is platform dependent.
     * @see javax.swing.UIManager#getSystemLookAndFeelClassName()
     */
    public static final String LAFChoiceNativeSystem = "swing.systemlaf";

    /**
     * This look&amp;feel choice denotes the Swing look&amp;feel
     * that best approximates the system look&amp;feel, except that
     * the Gtk look&amp;feel is used instead of the cross-platform
     * look&amp;feel if SWT is based on Gtk and if it works fine.
     * (This typically affects Linux systems with KDE desktop.)
     * It is platform dependent.
     */
    public static final String LAFChoiceNativeSystemPreferGtk = "swing.systemlaf+gtk";

    /**
     * This look&amp;feel choice denotes the Swing look&amp;feel
     * that best approximates the system look&amp;feel, except that
     * the Gtk look&amp;feel is avoided.
     * (This typically affects Linux systems with Gtk desktop.)
     * It is platform dependent.
     */
    public static final String LAFChoiceNativeSystemNoGtk = "swing.systemlaf-gtk";

    private String lafChoice;
    // Set the default look&feel choice.
    {
        // On JDK 1.6, we have to avoid the Swing Gtk look&feel, to work around
        // https://bugs.eclipse.org/bugs/show_bug.cgi?id=126931
        if (Platform.JAVA_VERSION >= Platform.javaVersion(1, 6, 0)) {
            lafChoice = LAFChoiceNativeSystemNoGtk;
        } else {
            // Set the default to LAFChoiceNativeSystemPreferGtk, so that on
            // Unix systems with a GNOME desktop and with themes supported by
            // the Swing Gtk look&feel this Gtk look&feel is used; it resembles
            // the system look&feel more closely than Metal or Nimbus.
            lafChoice = LAFChoiceNativeSystemPreferGtk;
        }
    }

    /**
     * Returns the look&amp;feel choice. It can be a class name or one
     * the shorthands defined in this class.
     */
    public String getLAFChoice() {
        return lafChoice;
    }

    /**
     * Specifies which look&amp;feel should be used.
     * @param choice Either a class name, such as
     *               <code>"javax.swing.plaf.metal.MetalLookAndFeel"</code>,
     *               <code>"com.sun.java.swing.plaf.windows.WindowsLookAndFeel"</code>,
     *               <code>"com.sun.java.swing.plaf.gtk.GTKLookAndFeel"</code>,
     *               <code>"com.sun.java.swing.plaf.motif.MotifLookAndFeel"</code>,
     *               <code>"apple.laf.AquaLookAndFeel"</code>,
     *               or one of the shorthands
     *               {@link #LAFChoiceSwingDefault},
     *               {@link #LAFChoiceCrossPlatform},
     *               {@link #LAFChoiceNativeSystem},
     *               {@link #LAFChoiceNativeSystemPreferGtk},
     *               {@link #LAFChoiceNativeSystemNoGtk}.
     */
    public void setLAFChoice(String choice) {
        lafChoice = choice;
    }

    // ========================================================================
    // Overridable API

    private static final String GTK_LOOK_AND_FEEL_NAME = "com.sun.java.swing.plaf.gtk.GTKLookAndFeel"; //$NON-NLS-1$

    /**
     * Sets the desired look&amp;feel.
     * @see #setLAFChoice(String)
     */
    public void setLookAndFeel()
        throws ClassNotFoundException,
               InstantiationException,
               IllegalAccessException,
               UnsupportedLookAndFeelException {
        assert EventQueue.isDispatchThread(); // On AWT event thread

        // If the user has specified the Swing look&feel through the
        // system property, and the application has also specified the
        // look&feel, obey the user. The user is always right.
        if (System.getProperty("swing.defaultlaf") != null)
            return;

        String laf = getLAFChoice();
        if (LAFChoiceSwingDefault.equals(laf)) {
            return;
        } else if (LAFChoiceCrossPlatform.equals(laf)) {
            laf = UIManager.getCrossPlatformLookAndFeelClassName();
        } else if (LAFChoiceNativeSystem.equals(laf)) {
            laf = UIManager.getSystemLookAndFeelClassName();
        } else if (LAFChoiceNativeSystemPreferGtk.equals(laf)) {
            laf = UIManager.getSystemLookAndFeelClassName();
            if (Platform.isGtk()
                && laf.equals(UIManager.getCrossPlatformLookAndFeelClassName())) {
                
                laf = GTK_LOOK_AND_FEEL_NAME;
            }
            if (laf.equals(GTK_LOOK_AND_FEEL_NAME)) {
                // Try the Gtk look&feel. 
                try {
                    doSetLookAndFeel(GTK_LOOK_AND_FEEL_NAME);
                    return;
                } catch (ClassNotFoundException e) {
                } catch (InstantiationException e) {
                } catch (IllegalAccessException e) {
                } catch (UnsupportedLookAndFeelException e) {
                }
                // Second try: Use cross platform look and feel
                laf = UIManager.getCrossPlatformLookAndFeelClassName();
            }
        } else if (LAFChoiceNativeSystemNoGtk.equals(laf)) {
            laf = UIManager.getSystemLookAndFeelClassName();
            if (GTK_LOOK_AND_FEEL_NAME.equals(laf))
                laf = UIManager.getCrossPlatformLookAndFeelClassName();
        }

        doSetLookAndFeel(laf);
    }

    private static void doSetLookAndFeel(String laf)
        throws ClassNotFoundException,
               InstantiationException,
               IllegalAccessException,
               UnsupportedLookAndFeelException {
        // Although UIManager.setLookAndFeel is specified to throw an
        // UnsupportedLookAndFeelException when the look&feel is not supported,
        // the com.sun.java.swing.plaf.gtk.GTKParser of JDK 1.5.0 reports
        // failures by doing System.err.println() and instead sets an unthemed
        // Gtk look&feel, which is very ugly.
        // Typically this happens on Linux systems with KDE desktop. The
        // $HOME/.gtkrc-2.0 file created by the KDE control center refers to
        // /opt/gnome/share/themes/Qt/gtk-2.0/gtkrc, which specifies a theming
        // engine "qtengine", which exists in C++ code but not in the
        // com.sun.java.swing.plaf.gtk mockup. The error message in this case
        // reads:
        // "/opt/gnome/share/themes/Qt/gtk-2.0/gtkrc:5: Engine "qtengine" is unsupported, ignoring"
        //
        // This can also happens even when running the GNOME desktop, when certain themes are selected.  
        PrintStream origSystemErr = System.err;
        try {
            System.setErr(
                new PrintStream(origSystemErr) {
                    public void print(String s) {
                        throw new UnsupportedLookAndFeelRuntimeException(s);
                    }
                    public void println(String s) {
                        throw new UnsupportedLookAndFeelRuntimeException(s);
                    }
                });
            UIManager.setLookAndFeel(laf);
        } catch (UnsupportedLookAndFeelRuntimeException e) {
            UnsupportedLookAndFeelException newExc =
                new UnsupportedLookAndFeelException(e.getMessage());
            newExc.initCause(e);
            throw newExc;
        } finally {
            System.setErr(origSystemErr);
        }
    }

    /**
     * This exception signals an unsupported look&feel.
     */
    static class UnsupportedLookAndFeelRuntimeException extends RuntimeException {
        UnsupportedLookAndFeelRuntimeException(String message) {
            super(message);
        }
    }

    // ========================================================================
    // Singleton design pattern

    private static LookAndFeelHandler theHandler = new LookAndFeelHandler();

    /**
     * Returns the currently active singleton of this class.
     */
    public static LookAndFeelHandler getInstance() {
        return theHandler;
    }

    /**
     * Replaces the singleton of this class.
     * @param instance An instance of this class or of a customized subclass.
     */
    public static void setInstance(LookAndFeelHandler instance) {
        theHandler = instance;
    }

}
