/*******************************************************************************
 * Copyright (c) 2007 SAS Institute Inc., ILOG S.A.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     SAS Institute Inc. - initial API and implementation
 *     ILOG S.A. - initial API and implementation
 *******************************************************************************/
package org.eclipse.albireo.internal;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.EventQueue;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.AWTEventListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.albireo.core.ThreadingHandler;
import org.eclipse.swt.widgets.Display;

/**
 * A listener that insures the proper modal behavior of Swing dialogs when running
 * within a SWT environment. When initialized, it blocks and unblocks SWT input
 * as modal Swing dialogs are shown and hidden. 
 */
public class AwtDialogListener implements AWTEventListener, ComponentListener {
    
    // modalDialogs should be accessed only from the AWT thread, so no
    // synchronization is needed. 
    private final List modalDialogs = new ArrayList();
    private final Display display;
    
    /**
     * Registers this object as an AWT event listener so that Swing dialogs have the 
     * proper modal behavior in the containing SWT environment. This is called automatically
     * when you construct a {@link EmbeddedSwingComposite}, and it
     * need not be called separately in that case.  
     * @param shell 
     */
    public AwtDialogListener(Display display) {
        assert display != null;
        
        this.display = display;
        Toolkit.getDefaultToolkit().addAWTEventListener(this, AWTEvent.WINDOW_EVENT_MASK);
    }
    
    private void handleRemovedDialog(Dialog awtDialog, boolean removeListener) {
        assert awtDialog != null;
        assert modalDialogs != null;
        assert display != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        // System.out.println("Remove dialog: " + awtDialog);
        if (removeListener) {
            awtDialog.removeComponentListener(this);
        }
        // Note: there is no isModal() check here because the dialog might 
        // have been changed from modal to non-modal after it was opened. In this case
        // the currently visible dialog would still act modal and we'd need to unblock
        // SWT here when it goes away.
        if (modalDialogs.remove(awtDialog)) {
            ThreadingHandler.getInstance().asyncExec(display, new Runnable() {
                public void run() {
                    SwtInputBlocker.unblock();
                }
            });            
        }
    }

    private void handleAddedDialog(final Dialog awtDialog) {
        assert awtDialog != null;
        assert modalDialogs != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        // System.out.println("Add dialog: " + awtDialog);
        
        // Don't block if 
        // 1) the the dialog has already triggered a block
        // 2) the dialog is not modal, or
        // 3) the dialog is not (yet?) visible
        // It's not clear why/when case 3 would happen, but it has been reported and the consequences
        // are severe if the check is not in place. 
        if (modalDialogs.contains(awtDialog) || !awtDialog.isModal() || !awtDialog.isVisible()) {
            return;
        }
        modalDialogs.add(awtDialog);
        awtDialog.addComponentListener(this);
        ThreadingHandler.getInstance().asyncExec(display, new Runnable() {
            public void run() {
                SwtInputBlocker.block(AwtDialogListener.this);
            }
        });        
    }
    
    void requestFocus() {
        // TODO: in early testing this did not always bring the dialog to the top 
        // under some Linux desktops/window managers (e.g. metacity under GNOME).
        // Re-test with recent changes
        EventQueue.invokeLater(new Runnable() {
            public void run() {
                assert modalDialogs != null;
                
                int size = modalDialogs.size();
                if (size > 0) {
                    final Dialog awtDialog = (Dialog)modalDialogs.get(size - 1);

                    // In one case, a call to requestFocus() alone does not 
                    // bring the AWT dialog to the top. This happens if the 
                    // dialog is given a null parent frame. When opened, the dialog
                    // can be hidden by the SWT window even when it obtains focus.
                    // Calling toFront() solves the problem.
                    
                    // System.out.println("Bringing to front");

                    Component focusOwner = awtDialog.getMostRecentFocusOwner();
                    if (focusOwner == null) {
                        focusOwner = awtDialog; // try the dialog itself in this case
                    }
                    try {
                        focusOwner.requestFocus();
                        awtDialog.toFront();
                    } catch (NullPointerException e) {
                        // Some dialogs (e.g. Windows page setup and print dialogs on JDK 1.5+) throw an NPE on
                        // requestFocus(). There's no way to check ahead of time, so just swallow the NPE here. 
                    }
                }
            }
        });
    }

    private void handleOpenedWindow(WindowEvent event) {
        assert event != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        Window window = event.getWindow();
        if (window instanceof Dialog) {
            handleAddedDialog((Dialog)window);
        }
    }
    
    private void handleClosedWindow(WindowEvent event) {
        assert event != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        // Dispose-based close
        Window window = event.getWindow();
        if (window instanceof Dialog) {
            // Remove dialog and component listener
            handleRemovedDialog((Dialog)window, true);
        }
    }

    private void handleClosingWindow(WindowEvent event) {
        assert event != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        // System-based close 
        Window window = event.getWindow();
        if (window instanceof Dialog) {
            final Dialog dialog = (Dialog) window;
            // Defer until later. Bad things happen if 
            // handleRemovedDialog() is called directly from 
            // this event handler. The Swing dialog does not close
            // properly and its modality remains in effect.
            EventQueue.invokeLater(new Runnable() {
                public void run() {
                    // Remove dialog and component listener
                    handleRemovedDialog(dialog, true);
                }
            });
        }
    }
    
    public void dispose() {
        Toolkit.getDefaultToolkit().removeAWTEventListener(this);
    }
    
    public void eventDispatched(AWTEvent event) {
        assert event != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        switch (event.getID()) {
        case WindowEvent.WINDOW_OPENED:
            handleOpenedWindow((WindowEvent)event);
            break;
            
        case WindowEvent.WINDOW_CLOSED:
            handleClosedWindow((WindowEvent)event);
            break;

        case WindowEvent.WINDOW_CLOSING:
            handleClosingWindow((WindowEvent)event);
            break;

        default:
            break;
        }
    }

    public void componentHidden(ComponentEvent e) {
        assert e != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        // System.out.println("Component hidden");
        Object obj = e.getSource();
        if (obj instanceof Dialog) {
            // Remove dialog but keep listener in place so that we know if/when it is set visible
            handleRemovedDialog((Dialog)obj, false);
        }
    }

    public void componentShown(ComponentEvent e) {
        assert e != null;
        assert EventQueue.isDispatchThread();    // On AWT event thread
        
        // System.out.println("Component shown");
        Object obj = e.getSource();
        if (obj instanceof Dialog) {
            handleAddedDialog((Dialog)obj);
        }
    }

    public void componentResized(ComponentEvent e) {
        // Ignore event
    }

    public void componentMoved(ComponentEvent e) {
        // Ignore event
    }
        
}
